@@ -5,19 +5,11 @@ from configparser import ConfigParser
 from datetime import datetime
 import importlib
 
-import stsci_rtd_theme
 if sys.version_info < (3, 11):
     import tomli as tomllib
 else:
     import tomllib
 
-def setup(app):
-    try:
-        app.add_css_file("stsci.css")
-    except AttributeError:
-        app.add_stylesheet("stsci.css")
-
-
 REPO_ROOT = Path(__file__).parent.parent.parent
 
 # Modules that automodapi will document need to be available
@@ -37,6 +29,7 @@ release = package.__version__
 
 extensions = [
     "sphinx_automodapi.automodapi",
+    "sphinxcontrib.jquery",
     "numpydoc",
     'sphinx_asdf',
 ]
@@ -45,11 +38,12 @@ autosummary_generate = True
 numpydoc_show_class_members = False
 autoclass_content = "both"
 
-html_theme = "stsci_rtd_theme"
+html_logo = 'static/stsci_pri_combo_mark_white.png'
+html_theme = "sphinx_rtd_theme"
 html_theme_options = {
-    "collapse_navigation": True
+    "collapse_navigation": True,
+    "sticky_navigation": False,
 }
-html_theme_path = [stsci_rtd_theme.get_html_theme_path()]
 html_domain_indices = True
 html_sidebars = {"**": ["globaltoc.html", "relations.html", "searchbox.html"]}
 html_use_index = True
