@@ -3,7 +3,6 @@ from functools import partial
 import hashlib
 import inspect
 import io
-from pkg_resources import parse_version
 import re
 import warnings
 import weakref
@@ -20,7 +19,6 @@ from asdf import treeutil
 from asdf.util import HashableDict
 from asdf import tagged
 from asdf import generic_io
-from jsonschema import validators
 
 from . import properties
 from . import schema as mschema
@@ -37,10 +35,10 @@ __all__ = ['to_fits', 'from_fits', 'fits_hdu_name', 'get_hdu', 'is_builtin_fits_
 
 _ASDF_EXTENSION_NAME = "ASDF"
 _FITS_SOURCE_PREFIX = "fits:"
-_NDARRAY_TAG = "tag:stsci.edu:asdf/core/ndarray-1.0.0"
-
-_ASDF_GE_2_6 = parse_version(asdf.__version__) >= parse_version('2.6')
-
+if asdf.versioning.default_version > "1.5.0":
+    _NDARRAY_TAG = "tag:stsci.edu:asdf/core/ndarray-1.1.0"
+else:
+    _NDARRAY_TAG = "tag:stsci.edu:asdf/core/ndarray-1.0.0"
 
 _builtin_regexes = [
     '', 'NAXIS[0-9]{0,3}', 'BITPIX', 'XTENSION', 'PCOUNT', 'GCOUNT',
@@ -133,8 +131,10 @@ def _get_hdu_pair(hdu_name, index=None):
     return pair
 
 
-def get_hdu(hdulist, hdu_name, index=None):
+def get_hdu(hdulist, hdu_name, index=None, _cache=None):
     pair = _get_hdu_pair(hdu_name, index=index)
+    if _cache is not None and pair in _cache:
+        return _cache[pair]
     try:
         hdu = hdulist[pair]
     except (KeyError, IndexError, AttributeError):
@@ -158,6 +158,8 @@ def get_hdu(hdulist, hdu_name, index=None):
                 "{0!r} HDU".format(
                     pair))
 
+    if _cache is not None:
+        _cache[pair] = hdu
     return hdu
 
 
@@ -274,6 +276,8 @@ def _fits_array_writer(fits_context, validator, _, instance, schema):
     if instance is None:
         return
 
+    instance_id = id(instance)
+
     instance = np.asanyarray(instance)
 
     if not len(instance.shape):
@@ -297,6 +301,10 @@ def _fits_array_writer(fits_context, validator, _, instance, schema):
                            index=index, hdu_type=hdu_type)
 
     hdu.data = instance
+    if instance_id in fits_context.extension_array_links:
+        if fits_context.extension_array_links[instance_id]() is not hdu:
+            raise ValueError("Linking one array to multiple hdus is not supported")
+    fits_context.extension_array_links[instance_id] = weakref.ref(hdu)
     hdu.ver = index + 1
 
 
@@ -323,7 +331,8 @@ def _fits_item_recurse(fits_context, validator, items, instance, schema):
 def _fits_type(fits_context, validator, items, instance, schema):
     if instance in ('N/A', '#TODO', '', None):
         return
-    return validators.Draft4Validator.VALIDATORS["type"](validator, items, instance, schema)
+
+    return asdf_schema.YAML_VALIDATORS["type"](validator, items, instance, schema)
 
 
 class FitsContext:
@@ -331,6 +340,7 @@ class FitsContext:
         self.hdulist = weakref.ref(hdulist)
         self.comment_stack = []
         self.sequence_index = None
+        self.extension_array_links = {}
 
 
 def _get_validators(hdulist):
@@ -350,11 +360,11 @@ def _get_validators(hdulist):
         'type': partial(_fits_type, fits_context),
     })
 
-    return validators
+    return validators, fits_context
 
 
 def _save_from_schema(hdulist, tree, schema):
-    def datetime_callback(node, json_id):
+    def datetime_callback(node):
         if isinstance(node, datetime.datetime):
             node = time.Time(node)
 
@@ -365,29 +375,32 @@ def _save_from_schema(hdulist, tree, schema):
 
     tree = treeutil.walk_and_modify(tree, datetime_callback)
 
-    if _ASDF_GE_2_6:
-        kwargs = {"_visit_repeat_nodes": True}
-    else:
-        kwargs = {}
+    kwargs = {"_visit_repeat_nodes": True}
 
-    validator = asdf_schema.get_validator(
-        schema, None, _get_validators(hdulist), **kwargs)
+    validators, context = _get_validators(hdulist)
+    validator = asdf_schema.get_validator(schema, None, validators, **kwargs)
 
     # This actually kicks off the saving
     validator.validate(tree, _schema=schema)
 
-    # Replace arrays in the tree that are identical to HDU arrays
-    # with ndarray-1.0.0 tagged objects with special source values
-    # that represent links to the surrounding FITS file.
-    def ndarray_callback(node, json_id):
-        if (isinstance(node, (np.ndarray, NDArrayType))):
+    # Now link extensions to items in the tree
+
+    def callback(node):
+        if id(node) in context.extension_array_links:
+            hdu = context.extension_array_links[id(node)]()
+            return _create_tagged_dict_for_fits_array(hdu, hdulist.index(hdu))
+        elif isinstance(node, (np.ndarray, NDArrayType)):
+            # in addition to links generated during validation
+            # replace arrays in the tree that are identical to HDU arrays
+            # with ndarray-1.0.0 tagged objects with special source values
+            # that represent links to the surrounding FITS file.
+            # This is important for general ASDF-in-FITS support
             for hdu_index, hdu in enumerate(hdulist):
                 if hdu.data is not None and node is hdu.data:
                     return _create_tagged_dict_for_fits_array(hdu, hdu_index)
-
         return node
 
-    tree = treeutil.walk_and_modify(tree, ndarray_callback)
+    tree = treeutil.walk_and_modify(tree, callback)
 
     return tree
 
@@ -426,7 +439,7 @@ def _normalize_arrays(tree):
     don't want the asdf library to notice the change in memory
     layout and duplicate the array in the embedded ASDF.
     """
-    def normalize_array(node, json_id):
+    def normalize_array(node):
         if isinstance(node, np.ndarray):
             # We can't use np.ascontiguousarray because it converts FITS_rec
             # to vanilla np.ndarray, which results in misinterpretation of
@@ -498,7 +511,12 @@ def to_fits(tree, schema, hdulist=None):
 
 def _create_asdf_hdu(tree):
     buffer = io.BytesIO()
-    asdf.AsdfFile(tree).write_to(buffer)
+    # convert all FITS_rec instances to numpy arrays, this is needed as
+    # some arrays loaded from the fits data for old files may not be defined
+    # in the current schemas. These will be loaded as FITS_rec instances but
+    # not linked back (and safely converted) on write if they are removed
+    # from the schema.
+    asdf.AsdfFile(util.convert_fitsrec_to_array_in_tree(tree)).write_to(buffer)
     buffer.seek(0)
 
     data = np.array(buffer.getbuffer(), dtype=np.uint8)[None, :]
@@ -511,11 +529,11 @@ def _create_asdf_hdu(tree):
 # READER
 
 
-def _fits_keyword_loader(hdulist, fits_keyword, schema, hdu_index, known_keywords):
+def _fits_keyword_loader(hdulist, fits_keyword, schema, hdu_index, known_keywords, fits_hdu_cache):
 
     hdu_name = _get_hdu_name(schema)
     try:
-        hdu = get_hdu(hdulist, hdu_name, hdu_index)
+        hdu = get_hdu(hdulist, hdu_name, hdu_index, _cache=fits_hdu_cache)
     except AttributeError:
         return None
 
@@ -533,16 +551,16 @@ def _fits_keyword_loader(hdulist, fits_keyword, schema, hdu_index, known_keyword
     return val
 
 
-def _fits_array_loader(hdulist, schema, hdu_index, known_datas, context):
+def _fits_array_loader(hdulist, schema, hdu_index, known_datas, fits_hdu_cache):
     hdu_name = _get_hdu_name(schema)
     _assert_non_primary_hdu(hdu_name)
     try:
-        hdu = get_hdu(hdulist, hdu_name, hdu_index)
+        hdu = get_hdu(hdulist, hdu_name, hdu_index, _cache=fits_hdu_cache)
     except AttributeError:
         return None
 
     known_datas.add(hdu)
-    return from_fits_hdu(hdu, schema, context._cast_fits_arrays)
+    return from_fits_hdu(hdu, schema)
 
 
 def _schema_has_fits_hdu(schema):
@@ -571,13 +589,19 @@ def _load_from_schema(hdulist, schema, tree, context, skip_fits_update=False):
     # This is needed to constrain the loop over HDU's when resolving arrays.
     max_extver = max(hdu.ver for hdu in hdulist) if len(hdulist) else 0
 
+    # hdulist.__getitem__ is surprisingly slow (2 ms per call on my system
+    # for a nirspec mos file with ~500 extensions) so we use a cache
+    # here to handle repeated accesses. A lru_cache around get_hdu
+    # was not used as hdulist is not hashable.
+    hdu_cache = {}
+
     def callback(schema, path, combiner, ctx, recurse):
         result = None
         if not skip_fits_update and 'fits_keyword' in schema:
             fits_keyword = schema['fits_keyword']
             result = _fits_keyword_loader(
                 hdulist, fits_keyword, schema,
-                ctx.get('hdu_index'), known_keywords)
+                ctx.get('hdu_index'), known_keywords, hdu_cache)
 
             if result is None and context._validate_on_assignment:
                 validate.value_change(path, result, schema, context)
@@ -591,7 +615,7 @@ def _load_from_schema(hdulist, schema, tree, context, skip_fits_update=False):
         elif 'fits_hdu' in schema and (
                 'max_ndim' in schema or 'ndim' in schema or 'datatype' in schema):
             result = _fits_array_loader(
-                hdulist, schema, ctx.get('hdu_index'), known_datas, context)
+                hdulist, schema, ctx.get('hdu_index'), known_datas, hdu_cache)
 
             if result is None and context._validate_on_assignment:
                 validate.value_change(path, result, schema, context)
@@ -674,6 +698,7 @@ def from_fits(hdulist, schema, context, skip_fits_update=None, **kwargs):
         The `DataModel` to update
 
     skip_fits_update : bool or None
+        DEPRECATED
         When `False`, models opened from FITS files will proceed
         and load the FITS header values into the model.
         When `True` and the FITS file has an ASDF extension, the
@@ -704,7 +729,6 @@ def from_fits(hdulist, schema, context, skip_fits_update=None, **kwargs):
 
 
 def from_fits_asdf(hdulist,
-                   ignore_version_mismatch=True,
                    ignore_unrecognized_tag=False,
                    **kwargs):
     """
@@ -717,7 +741,6 @@ def from_fits_asdf(hdulist,
     except (KeyError, IndexError, AttributeError):
         # This means there is no ASDF extension
         return asdf.AsdfFile(
-            ignore_version_mismatch=ignore_version_mismatch,
             ignore_unrecognized_tag=ignore_unrecognized_tag,
         )
 
@@ -729,7 +752,6 @@ def from_fits_asdf(hdulist,
     }
     af = asdf.open(
         generic_file,
-        ignore_version_mismatch=ignore_version_mismatch,
         ignore_unrecognized_tag=ignore_unrecognized_tag,
         ignore_missing_extensions=ignore_missing_extensions,
         **akwargs
@@ -740,7 +762,7 @@ def from_fits_asdf(hdulist,
 
 
 def _map_hdulist_to_arrays(hdulist, af):
-    def callback(node, json_id):
+    def callback(node):
         if (
                 isinstance(node, NDArrayType) and
                 isinstance(node._source, str) and
@@ -762,34 +784,28 @@ def _map_hdulist_to_arrays(hdulist, af):
             data = hdulist[pair].data
             return data
         return node
-    af.tree = treeutil.walk_and_modify(af.tree, callback)
+    # don't assign to af.tree to avoid an extra validation
+    af._tree = treeutil.walk_and_modify(af.tree, callback)
 
 
-def from_fits_hdu(hdu, schema, cast_arrays=True):
+def from_fits_hdu(hdu, schema):
     """
     Read the data from a fits hdu into a numpy ndarray
     """
     data = hdu.data
 
-    if cast_arrays:
-        # Save the column listeners for possible restoration
-        if hasattr(data, '_coldefs'):
-            listeners = data._coldefs._listeners
-        else:
-            listeners = None
-
-        # Cast array to type mentioned in schema
-        data = properties._cast(data, schema)
-
-        # Casting a table loses the column listeners, so restore them
-        if listeners is not None:
-            data._coldefs._listeners = listeners
+    # Save the column listeners for possible restoration
+    if hasattr(data, '_coldefs'):
+        listeners = data._coldefs._listeners
     else:
-        # Correct the pseudo-unsigned int problem (this normally occurs
-        # inside properties._cast, but we still need to do it even
-        # when not casting, otherwise arrays from FITS will fail validation).
-        if isinstance(data, fits.FITS_rec):
-            data.dtype = util.rebuild_fits_rec_dtype(data)
+        listeners = None
+
+    # Cast array to type mentioned in schema
+    data = properties._cast(data, schema)
+
+    # Casting a table loses the column listeners, so restore them
+    if listeners is not None:
+        data._coldefs._listeners = listeners
 
     return data
 
@@ -821,6 +837,9 @@ def _verify_skip_fits_update(skip_fits_update, hdulist, asdf_struct, context):
     """
     if skip_fits_update is None:
         skip_fits_update = util.get_envar_as_boolean('SKIP_FITS_UPDATE', None)
+    if skip_fits_update is not None:
+        # warn if the value was not None (defined by the user)
+        warnings.warn("skip_fits_update is deprecated and will be removed", DeprecationWarning)
 
     # If skipping has been explicitly disallowed, indicate as such.
     if skip_fits_update is False:
