@@ -2,12 +2,7 @@
 This module supports the entry points for ASDF support for the `jwst.datamodels`.
 """
 
-import sys
-
-if sys.version_info < (3, 9):
-    import importlib_resources
-else:
-    import importlib.resources as importlib_resources
+import importlib.resources
 
 
 from asdf.resource import DirectoryResourceMapping
@@ -26,7 +21,7 @@ def get_resource_mappings():
     list of the `asdf.resource.ResourceMapping` instances containing the `jwst.datamodels`
     schemas.
     """
-    resources_root = importlib_resources.files(datamodels)
+    resources_root = importlib.resources.files(datamodels)
     if not resources_root.is_dir():
         raise RuntimeError(f"Missing resources directory: {resources_root=}")
 
