@@ -1,10 +1,9 @@
 from astropy.time import Time
+
 from stdatamodels import DataModel as _DataModel
 
-# from ..lib.basic_utils import deprecate_class
 
-
-__all__ = ["DataModel", "JwstDataModel"]
+__all__ = ["JwstDataModel"]
 
 
 class JwstDataModel(_DataModel):
@@ -53,9 +52,3 @@ class JwstDataModel(_DataModel):
         super().on_save(init)
 
         self.meta.date = Time.now().isot
-
-
-# We may want to deprecate DataModel
-# @deprecate_class(JwstDataModel)
-class DataModel(JwstDataModel):
-    pass
