@@ -1,5 +1,5 @@
 
-from stcal.dynamicdq import dynamic_mask
+from stdatamodels.dynamicdq import dynamic_mask
 from .dqflags import pixel
 from .reference import ReferenceFileModel
 
