@@ -1,21 +1,25 @@
+import contextlib
 import os
+import warnings
 
+from asdf.exceptions import ValidationError
 from astropy.io import fits
 from astropy.table import Table
 from astropy.time import Time
+from numpy.lib.recfunctions import merge_arrays
 from numpy.testing import assert_allclose, assert_array_equal
 import numpy as np
+from numpy.lib.recfunctions import drop_fields
 import pytest
-import warnings
 
 from stdatamodels.jwst.datamodels import (JwstDataModel, ImageModel, MaskModel, AsnModel,
                                           MultiSlitModel, SlitModel,
-                                          DrizProductModel, MultiProductModel, MIRIRampModel,
-                                          SlitDataModel, IFUImageModel, ABVegaOffsetModel)
+                                          NirspecFlatModel, NirspecQuadFlatModel,
+                                          SlitDataModel, IFUImageModel, ABVegaOffsetModel,
+                                          Level1bModel)
 from stdatamodels.jwst import datamodels
 from stdatamodels.jwst.datamodels import _defined_models as defined_models
 
-
 ROOT_DIR = os.path.join(os.path.dirname(__file__), 'data')
 FITS_FILE = os.path.join(ROOT_DIR, 'test.fits')
 ASN_FILE = os.path.join(ROOT_DIR, 'association.json')
@@ -51,10 +55,9 @@ def test_init_from_pathlib(tmp_path):
     path = tmp_path / "pathlib.fits"
     model1 = datamodels.ImageModel((50, 50))
     model1.save(path)
-    model = datamodels.open(path)
-
-    # Test is basically, did we open the model?
-    assert isinstance(model, ImageModel)
+    with datamodels.open(path) as model:
+        # Test is basically, did we open the model?
+        assert isinstance(model, ImageModel)
 
 
 @pytest.mark.parametrize('which_file, skip_fits_update, expected_exp_type',
@@ -77,23 +80,30 @@ def test_skip_fits_update(jail_environ,
     """Test skip_fits_update setting"""
     # Setup the FITS file, modifying a header value
     path = make_models[which_file]
-    hduls = fits.open(path)
-    hduls[0].header['exp_type'] = 'FGS_DARK'
+    with fits.open(path) as hduls:
+        hduls[0].header['exp_type'] = 'FGS_DARK'
+
+        # Decide how to skip. If using the environmental,
+        # set that and pass None to the open function.
+        try:
+            del os.environ['SKIP_FITS_UPDATE']
+        except KeyError:
+            # No need to worry, environmental doesn't exist anyways
+            pass
 
-    # Decide how to skip. If using the environmental,
-    # set that and pass None to the open function.
-    try:
-        del os.environ['SKIP_FITS_UPDATE']
-    except KeyError:
-        # No need to worry, environmental doesn't exist anyways
-        pass
-    if use_env:
         if skip_fits_update is not None:
-            os.environ['SKIP_FITS_UPDATE'] = str(skip_fits_update)
-            skip_fits_update = None
+            ctx = pytest.warns(DeprecationWarning, match="skip_fits_update is deprecated")
+        else:
+            ctx = contextlib.nullcontext()
 
-    with datamodels.open(hduls, skip_fits_update=skip_fits_update) as model:
-        assert model.meta.exposure.type == expected_exp_type
+        if use_env:
+            if skip_fits_update is not None:
+                os.environ['SKIP_FITS_UPDATE'] = str(skip_fits_update)
+                skip_fits_update = None
+
+        with ctx:
+            with datamodels.open(hduls, skip_fits_update=skip_fits_update) as model:
+                assert model.meta.exposure.type == expected_exp_type
 
 
 def test_asnmodel_table_size_zero():
@@ -312,11 +322,7 @@ def test_all_datamodels_init(model):
     """
     Test that all current datamodels can be initialized.
     """
-    if model in (DrizProductModel, MultiProductModel, MIRIRampModel):
-        with pytest.warns(DeprecationWarning):
-            model()
-    else:
-        model()
+    model()
 
 
 def test_meta_date_management(tmp_path):
@@ -376,6 +382,139 @@ def test_ramp_model_zero_frame_by_dimensions():
         assert ramp.zeroframe.shape == zdims
 
 
+@pytest.fixture
+def oifits_ami_model():
+    m = datamodels.AmiOIModel()
+    m.meta.telescope = 'JWST'
+    m.meta.origin = 'STScI'
+    m.meta.instrument.name = 'NIRISS'
+    m.meta.program.pi_name = 'UNKNOWN'
+    m.meta.target.proposer_name = 'AB DOR'
+    m.meta.observation.date = '2022-06-05'
+    m.meta.oifits.array_name = 'g7s6'
+    m.meta.oifits.instrument_mode = 'NRM'
+
+    m.array = [
+        (
+            'A1', 'A1', 1, 0.,
+            [ 2.64000000e+00, -1.61653377e-16,  0.00000000e+00], 5.04539835,
+            'RADIUS', [ 2.60973996, -0.39856915]
+        ),
+        (
+            'A2', 'A2', 2, 0.,
+            [ 1.39996111e-16,  2.28631000e+00,  0.00000000e+00], 5.04539835,
+            'RADIUS', [-0.34517145, -2.260104  ]
+        ),
+        (
+            'A3', 'A3', 3, 0.,
+            [ 1.32000010e+00, -2.28631000e+00,  0.00000000e+00], 5.04539835,
+            'RADIUS', [ 1.65004153,  2.06081941]
+        )
+    ]
+    m.target = [(
+        1, 'AB DOR', 82.18704833, -65.44869139, 2000., 0., 0., 0., 'UNKNOWN',
+        'OPTICAL', 29.15, 164.421, 0., 0., 65.3199, 0., 'K0V')]
+    m.t3 = [
+        (
+            1, 0., 59735., 0.3772, 0.71125417, 0.73752607, -1.81575411, 0.73752607,
+            1.86153485, 2.9549114, -4.32092341, -1.99521297, [1, 2, 3],  0
+        ),
+    ]
+    m.vis = [
+        (
+            1, 0., 59735., 0.3772, 0.84231787, 0.00669874, -10.57082496, 1.89098664,
+            1.86153485,  2.9549114 , [1, 2],  0
+        ),
+        (
+            1, 0., 59735., 0.3772, 0.91448467, 0.00621287, -22.26334374, 1.27637574,
+            -2.45938856,  0.95969843, [1, 3],  0
+        )
+    ]
+    m.vis2 = [
+        (1, 0., 59735., 0.3772, 0.70949939, 0.01131277,  1.86153485, 2.9549114 , [1, 2],  0),
+        (1, 0., 59735., 0.3772, 0.8362822 , 0.0113618 , -2.45938856, 0.95969843, [1, 3],  0)
+    ]
+    m.wavelength = [(4.817e-06, 2.98e-07)]
+    return m
+
+
+def test_amioi_model_oifits_validate(oifits_ami_model):
+    oifits_ami_model.validate()
+
+
+def test_amioi_model_oifits_update(oifits_ami_model):
+    oifits_ami_model.update({})
+
+
+def test_amioi_model_oifits_compliance(tmp_path, oifits_ami_model):
+    """
+    This test cannot fully test oifits compliance but the schema for the model
+    provides some checks. This test checks that the model in the oifits_ami_model
+    fixture provides a 'passable' OIFITS file.
+    """
+    fn = tmp_path / "test.fits"
+    oifits_ami_model.save(fn)
+
+
+@pytest.mark.parametrize('attr',
+                         [
+                             'meta.telescope',
+                             'meta.origin',
+                             'meta.instrument.name',
+                             'meta.program.pi_name',
+                             'meta.target.proposer_name',
+                             'meta.observation.date',
+                             'meta.oifits.array_name',
+                             'meta.oifits.instrument_mode',
+                             'array',
+                             'target',
+                             'wavelength',
+                         ])
+def test_amioi_model_oifits_keyword_validation(tmp_path, oifits_ami_model, attr):
+    """
+    Remove some critical keywords or tables and make sure the model fails to
+    save/validate This test cannot fully test oifits compliance but at least
+    does some sanity checks that a file produced using AmiOiModel produces a
+    file containing items that should make it oifits compliant
+    """
+    fn = tmp_path / "test.fits"
+
+    node = oifits_ami_model
+    *branches, leaf = attr.split('.')
+    for branch in branches:
+        node = getattr(node, branch)
+    delattr(node, leaf)
+
+    with pytest.raises(ValidationError):
+        oifits_ami_model.save(fn)
+
+
+@pytest.mark.parametrize('keep', ['vis', 'vis2', 't3'])
+def test_amioi_model_oifits_datatable(tmp_path, oifits_ami_model, keep):
+    fn = tmp_path / "test.fits"
+    for table in ('vis', 'vis2', 't3'):
+        if table == keep:
+            continue
+        delattr(oifits_ami_model, table)
+
+    # since we have 1 data table, this should pass
+    oifits_ami_model.save(fn)
+
+    # removing the data table should cause saving to fail
+    delattr(oifits_ami_model, keep)
+    with pytest.raises(ValidationError):
+        oifits_ami_model.save(fn)
+
+
+@pytest.mark.parametrize('table_name', ['array', 'target', 'vis', 'vis2', 't3', 'wavelength'])
+def test_amioi_model_oifits_extra_columns(tmp_path, oifits_ami_model, table_name):
+    table_data = getattr(oifits_ami_model, table_name)
+    new_table_data = merge_arrays([table_data, np.zeros(table_data.size, dtype=[('extra', 'f8')])], flatten=True)
+    setattr(oifits_ami_model, table_name, new_table_data)
+    fn = tmp_path / "test.fits"
+    oifits_ami_model.save(fn)
+
+
 def test_dq_def_roundtrip(tmp_path):
     """
     Open a MaskModel with a defined DQ array and dq_def that modifies the
@@ -410,3 +549,80 @@ def test_dq_def_roundtrip(tmp_path):
     assert diff is not None
     total_diff = sum(sum(diff))
     assert total_diff == 0
+
+
+@pytest.mark.parametrize("shape", [None, 10])
+@pytest.mark.parametrize("model", [NirspecFlatModel, NirspecQuadFlatModel])
+def test_nirspec_flat_table_migration(tmp_path, model, shape):
+    fn = tmp_path / 'test.fits'
+
+    def make_data(table_dtype):
+        if shape:
+            fake_data = [('ABC', shape, [0.1] * shape, [2.0] * shape, [3.0] * shape)]
+            dtype = [(n, table_dtype[n], shape if n != 'slit_name' else ()) for n in table_dtype.fields]
+        else:
+            fake_data = [('ABC', 1, 0.1, 2.0, 3.0)]
+            dtype = table_dtype
+        return np.array(fake_data, dtype=dtype)
+
+    m = model()
+    if model == NirspecQuadFlatModel:
+        m.quadrants.append(m.quadrants.item())
+        m.quadrants[0].flat_table = make_data(m.quadrants[0].flat_table.dtype)
+    else:
+        m.flat_table = make_data(m.flat_table.dtype)
+    m.save(fn)
+    with fits.open(fn) as ff:
+        for ext in ff:
+            if ext.name != 'FAST_VARIATION':
+                continue
+            # drop the error column
+            ext.data = drop_fields(ext.data, 'error')
+        ff.writeto(fn, overwrite=True)
+
+    def check_error_column(model):
+        if isinstance(model, NirspecQuadFlatModel):
+            table = model.quadrants[0].flat_table
+        else:
+            table = model.flat_table
+        assert np.all(np.isnan(table['error']))
+
+    # check that migration works with datamodels.open
+    with datamodels.open(fn) as dm:
+        check_error_column(dm)
+    # and with DataModel(fn)
+    with model(fn) as dm:
+        check_error_column(dm)
+
+
+def test_moving_target_table_migration(tmp_path):
+    fn = tmp_path / 'test_mt.fits'
+
+    def make_data(table_dtype):
+        shape = 10
+        fake_data = [(['0' * 23] * shape,) + ([0.1] * shape,) * (len(table_dtype) - 1)]
+        dtype = [(n, table_dtype[n], shape) for n in table_dtype.fields]
+        return np.array(fake_data, dtype=dtype)
+
+    m = Level1bModel()
+    m.moving_target = make_data(m.moving_target.dtype)
+    m.save(fn)
+    with fits.open(fn) as ff:
+        for ext in ff:
+            if ext.name != 'MOVING_TARGET_POSITION':
+                continue
+            # drop the error column
+            ext.data = drop_fields(ext.data, ('mt_v2', 'mt_v3'))
+        ff.writeto(fn, overwrite=True)
+
+    def check_error_column(model):
+        table = model.moving_target
+        assert np.all(np.isnan(table['mt_v2']))
+        assert np.all(np.isnan(table['mt_v3']))
+
+    # check that migration works with datamodels.open
+    with datamodels.open(fn) as dm:
+        check_error_column(dm)
+    # and with DataModel(fn)
+    with Level1bModel(fn) as dm:
+        check_error_column(dm)
