@@ -8,7 +8,7 @@ from .converters.jwst_models import (Gwa2SlitConverter, Slit2MsaConverter, Logic
                                      CoordsConverter, V23ToSkyConverter)
 
 
-JWST_TRANSFORM_CONVERTERS = [
+_CONVERTERS = [
     CoordsConverter(),
     Gwa2SlitConverter(),
     Slit2MsaConverter(),
@@ -21,22 +21,25 @@ JWST_TRANSFORM_CONVERTERS = [
     NIRISSGrismDispersionConverter(),
     GratingEquationConverter(),
     SnellConverter(),
-    V23ToSkyConverter(),
 ]
 
 # The order here is important; asdf will prefer to use extensions
 # that occur earlier in the list.
-TRANSFORM_MANIFEST_URIS = [
-    "asdf://stsci.edu/jwst_pipeline/manifests/jwst_transforms-1.0.0",
-    "asdf://stsci.edu/jwst_pipeline/manifests/jwst_transforms-0.7.0"
-]
-
-
 TRANSFORM_EXTENSIONS = [
     ManifestExtension.from_uri(
-        uri,
+        "asdf://stsci.edu/jwst_pipeline/manifests/jwst_transforms-1.1.0",
         legacy_class_names=["jwst.transforms.jwextension.JWSTExtension"],
-        converters=JWST_TRANSFORM_CONVERTERS,
+        converters=_CONVERTERS,
+    ),
+    ManifestExtension.from_uri(
+        "asdf://stsci.edu/jwst_pipeline/manifests/jwst_transforms-1.0.0",
+        legacy_class_names=["jwst.transforms.jwextension.JWSTExtension"],
+        converters=_CONVERTERS,
+    ),
+    ManifestExtension.from_uri(
+        "asdf://stsci.edu/jwst_pipeline/manifests/jwst_transforms-0.7.0",
+        legacy_class_names=["jwst.transforms.jwextension.JWSTExtension"],
+        # 0.7.0 support v23tosky, register it's converter
+        converters=_CONVERTERS + [V23ToSkyConverter(),],
     )
-    for uri in TRANSFORM_MANIFEST_URIS
 ]
