@@ -25,8 +25,7 @@ from . import fits_support
 from . import properties
 from . import schema as mschema
 from . import validate
-from .util import get_envar_as_boolean, remove_none_from_tree
-from . import s3_utils
+from .util import convert_fitsrec_to_array_in_tree, get_envar_as_boolean, remove_none_from_tree
 
 from .history import HistoryList
 
@@ -64,7 +63,7 @@ class DataModel(properties.ObjectNode):
 
     def __init__(self, init=None, schema=None, memmap=False,
                  pass_invalid_values=None, strict_validation=None,
-                 validate_on_assignment=None, cast_fits_arrays=True,
+                 validate_on_assignment=None,
                  validate_arrays=False, ignore_missing_extensions=True, **kwargs):
         """
         Parameters
@@ -95,8 +94,7 @@ class DataModel(properties.ObjectNode):
             will be used.
 
         memmap : bool
-            Turn memmap of FITS file on or off.  (default: False).  Ignored for
-            ASDF files.
+            Turn memmap of FITS/ASDF file on or off.  (default: False).
 
         pass_invalid_values : bool or None
             If `True`, values that do not validate the schema
@@ -119,11 +117,6 @@ class DataModel(properties.ObjectNode):
             If 'False', schema validation occurs only once at the time of write.
             Validation errors generate warnings.
 
-        cast_fits_arrays : bool
-            If `True`, arrays will be cast to the dtype described by the schema
-            when read from a FITS file.
-            If `False`, arrays will be read without casting.
-
         validate_arrays : bool
             If `True`, arrays will be validated against ndim, max_ndim, and datatype
             validators in the schemas.
@@ -140,6 +133,7 @@ class DataModel(properties.ObjectNode):
             - FITS
 
               skip_fits_update - bool or None
+                  DEPRECATED
                   `True` to skip updating the ASDF tree from the FITS headers, if possible.
                   If `None`, value will be taken from the environmental SKIP_FITS_UPDATE.
                   Otherwise, the default value is `True`.
@@ -159,7 +153,6 @@ class DataModel(properties.ObjectNode):
         self._strict_validation = strict_validation
         self._ignore_missing_extensions = ignore_missing_extensions
         self._validate_on_assignment = validate_on_assignment
-        self._cast_fits_arrays = cast_fits_arrays
         self._validate_arrays = validate_arrays
 
         kwargs.update({'ignore_missing_extensions': ignore_missing_extensions})
@@ -221,6 +214,7 @@ class DataModel(properties.ObjectNode):
             asdffile = init
 
         elif isinstance(init, fits.HDUList):
+            init = self._migrate_hdulist(init)
             asdffile = fits_support.from_fits(init, self._schema, self._ctx,
                                               **kwargs)
 
@@ -232,20 +226,15 @@ class DataModel(properties.ObjectNode):
             file_type = filetype.check(init)
 
             if file_type == "fits":
-                if s3_utils.is_s3_uri(init):
-                    init_fitsopen = s3_utils.get_object(init)
-                    memmap = None
-                else:
-                    init_fitsopen = init
-
-                hdulist = fits.open(init_fitsopen, memmap=memmap)
+                hdulist = fits.open(init, memmap=memmap)
+                self._file_references.append(_FileReference(hdulist))
+                hdulist = self._migrate_hdulist(hdulist)
                 asdffile = fits_support.from_fits(
                     hdulist, self._schema, self._ctx, **kwargs
                 )
-                self._file_references.append(_FileReference(hdulist))
 
             elif file_type == "asdf":
-                asdffile = self.open_asdf(init=init, **kwargs)
+                asdffile = self.open_asdf(init=init, memmap=memmap, **kwargs)
 
             else:
                 # TODO handle json files as well
@@ -299,6 +288,34 @@ class DataModel(properties.ObjectNode):
         # Call hook that sets model properties
         self.on_init(init)
 
+    def _migrate_hdulist(self, hdulist):
+        """
+        Optional method that can migrate a hdulist for an old (incompatible)
+        format to a new format.
+
+        For example, say you have a file with a table that is missing a
+        column that is now required. This method can be used to add the
+        missing column to the old file.
+
+        This method is only called when a datamodel is constructed using a
+        HDUList or a string (not when the DataModel is cloned or created
+        from an existing DataModel).
+
+        The migration occurs prior to loading the ASDF extension.
+
+        Parameters
+        ----------
+        hdulist : HDUList
+            The opened hdulist that this method should possibly migrate.
+
+        Returns
+        -------
+        migrated_hdulist : HDUList
+            The migrated/updated hdulist (this can be identical to the
+            input hdulist if no migration is needed).
+        """
+        return hdulist
+
     @property
     def _ctx(self):
         # self._ctx is a property to avoid reference cycle that would be
@@ -383,7 +400,10 @@ class DataModel(properties.ObjectNode):
     def clone(target, source, deepcopy=False, memo=None):
         if deepcopy:
             instance = copy.deepcopy(source._instance, memo=memo)
-            target._asdf = AsdfFile(instance)
+            target._asdf = AsdfFile()
+            # assign to private '_tree' to avoid validation caused
+            # by either using AsdfFile(instance) or target._asdf = tree
+            target._asdf._tree = instance
             target._instance = instance
         else:
             target._asdf = source._asdf
@@ -393,7 +413,6 @@ class DataModel(properties.ObjectNode):
                 target._file_references.append(file_reference)
 
         target._shape = source._shape
-        target._ctx = target
         target._no_asdf_extension = source._no_asdf_extension
 
     def copy(self, memo=None):
@@ -537,22 +556,18 @@ class DataModel(properties.ObjectNode):
 
     @staticmethod
     def open_asdf(init=None,
-                  ignore_version_mismatch=True,
                   ignore_unrecognized_tag=False,
                   **kwargs):
         """
         Open an asdf object from a filename or create a new asdf object
         """
         if isinstance(init, str):
-            if s3_utils.is_s3_uri(init):
-                init = s3_utils.get_object(init)
             asdffile = asdf.open(init,
-                                 ignore_version_mismatch=ignore_version_mismatch,
-                                 ignore_unrecognized_tag=ignore_unrecognized_tag)
+                                 ignore_unrecognized_tag=ignore_unrecognized_tag,
+                                 **kwargs)
 
         else:
             asdffile = AsdfFile(init,
-                            ignore_version_mismatch=ignore_version_mismatch,
                             ignore_unrecognized_tag=ignore_unrecognized_tag
                             )
         return asdffile
@@ -595,10 +610,15 @@ class DataModel(properties.ObjectNode):
             `~asdf.AsdfFile.write_to`.
         """
         self.on_save(init)
-        self.validate()
-        asdffile = self.open_asdf(self._instance, **kwargs)
+        self.validate()  # required to trigger ValidationWarning
+        tree = convert_fitsrec_to_array_in_tree(self._instance)
+        # don't open_asdf(tree) as this will cause a second validation of the tree
+        # instead open an empty tree, then assign to the hidden '_tree'
+        asdffile = self.open_asdf(None, **kwargs)
+        asdffile._tree = tree
         asdffile.write_to(init, *args, **kwargs)
 
+
     @classmethod
     def from_fits(cls, init, schema=None, **kwargs):
         """
