@@ -1,7 +1,6 @@
 # Licensed under a 3-clause BSD style license - see LICENSE.rst
 
 import copy
-import warnings
 import numpy as np
 from collections.abc import Mapping
 from astropy.io import fits
@@ -44,6 +43,7 @@ def _cast(val, schema):
         if isinstance(val, ndarray.NDArrayType):
             val = val._make_array()
 
+        allow_extra_columns = False
         if (_is_struct_array_schema(schema) and len(val) and
             (_is_struct_array_precursor(val) or _is_struct_array(val))):
             # we are dealing with a structured array. Because we may
@@ -51,6 +51,9 @@ def _cast(val, schema):
             # schema here:
             schema = copy.deepcopy(schema)
 
+            if 'allow_extra_columns' in schema:
+                allow_extra_columns = schema['allow_extra_columns']
+
             for t, v in zip(schema['datatype'], val[0]):
                 if not isinstance(t, Mapping):
                     continue
@@ -79,10 +82,20 @@ def _cast(val, schema):
                     t['shape'] = shape
 
         dtype = ndarray.asdf_datatype_to_numpy_dtype(schema['datatype'])
-        val = util.gentle_asarray(val, dtype)
+
+        # save columns in case this is cast back to a fitsrec
+        if hasattr(val, 'columns'):
+            cols = val.columns
+        else:
+            cols = None
+        val = util.gentle_asarray(val, dtype, allow_extra_columns=allow_extra_columns)
 
         if dtype.fields is not None:
             val = _as_fitsrec(val)
+            if cols is not None:
+                for col in cols:
+                    if col.name in val.names and col.unit is not None:
+                        val.columns[col.name].unit = col.unit
 
     if 'ndim' in schema and len(val.shape) != schema['ndim']:
         raise ValueError(
@@ -177,7 +190,15 @@ def _make_default_array(attr, schema, ctx):
                 if ndim is None:
                     shape = primary_array.shape
                 else:
-                    shape = primary_array.shape[-ndim:]
+                    if attr == "zeroframe":
+                        dims = primary_array.shape
+                        if len(dims) != 4:
+                            error_msg = "To allocate ZEROFRAME, the primary array must "
+                            error_msg += f"have 4 dimensions, but has {len(dims)}."
+                            raise IndexError(error_msg)
+                        shape = (dims[0], dims[2], dims[3])
+                    else:
+                        shape = primary_array.shape[-ndim:]
             elif ndim is None:
                 shape = (0,)
             else:
@@ -275,53 +296,11 @@ class ObjectNode(Node):
         else:
             return self._instance == other
 
-    def _is_deprecated(self, attr):
-        return "deprecated_properties" in self._schema and attr in self._schema["deprecated_properties"]
-
-    def _warn_deprecated_attr(self, attr):
-        node = self
-        path = []
-        while node._parent is not None:
-            if not isinstance(node, ListNode):
-                path.append(node._name)
-            node = node._parent
-
-        path.reverse()
-
-        old_attr = ".".join(path + [attr])
-
-        property_path = self._schema["deprecated_properties"][attr]
-        parts = property_path.split("/")
-        for part in parts:
-            if part == "..":
-                path.pop()
-            else:
-                path.append(part)
-
-        new_attr = ".".join(path)
-
-        warnings.warn(f"Attribute '{old_attr}' has been deprecated.  Use '{new_attr}' instead.", DeprecationWarning)
-
-    def _get_deprecated(self, attr):
-        property_path = self._schema["deprecated_properties"][attr]
-        node = self
-        parts = property_path.split("/")
-        for part in parts[:-1]:
-            if part == "..":
-                node = node._parent
-            else:
-                node = getattr(node, part)
-        return node, parts[-1]
-
     def __getattr__(self, attr):
+
         if attr.startswith('_'):
             raise AttributeError('No attribute {0}'.format(attr))
 
-        if self._is_deprecated(attr):
-            self._warn_deprecated_attr(attr)
-            node, new_attr = self._get_deprecated(attr)
-            return getattr(node, new_attr)
-
         schema = _get_schema_for_property(self._schema, attr)
         try:
             val = self._instance[attr]
@@ -345,10 +324,6 @@ class ObjectNode(Node):
     def __setattr__(self, attr, val):
         if attr.startswith('_'):
             self.__dict__[attr] = val
-        elif self._is_deprecated(attr):
-            self._warn_deprecated_attr(attr)
-            node, new_attr = self._get_deprecated(attr)
-            setattr(node, new_attr, val)
         else:
             schema = _get_schema_for_property(self._schema, attr)
             if val is None:
@@ -365,10 +340,6 @@ class ObjectNode(Node):
     def __delattr__(self, attr):
         if attr.startswith('_'):
             del self.__dict__[attr]
-        elif self._is_deprecated(attr):
-            self._warn_deprecated_attr(attr)
-            node, new_attr = self._get_deprecated(attr)
-            delattr(node, new_attr)
         else:
             schema = _get_schema_for_property(self._schema, attr)
             if validate.value_change(attr, None, schema, self._ctx) or self._ctx._pass_invalid_values:
@@ -432,30 +403,6 @@ class ListNode(Node):
         if self._ctx._validate_on_assignment:
             self._validate()
 
-    def __getslice__(self, i, j):
-        if isinstance(self._schema['items'], list):
-            r = range(*(slice(i, j).indices(len(self._instance))))
-            schema_parts = [
-                _get_schema_for_index(self._schema, x) for x in r
-            ]
-        else:
-            schema_parts = self._schema['items']
-        schema = {'type': 'array', 'items': schema_parts}
-        return _make_node(self._name, self._instance[i:j], schema, self._ctx, self)
-
-    def __setslice__(self, i, j, other):
-        parts = _unmake_node(other)
-        parts = [_cast(x, _get_schema_for_index(self._schema, k))
-                 for (k, x) in enumerate(parts)]
-        self._instance[i:j] = _unmake_node(other)
-        if self._ctx._validate_on_assignment:
-            self._validate()
-
-    def __delslice__(self, i, j):
-        del self._instance[i:j]
-        if self._ctx._validate_on_assignment:
-            self._validate()
-
     def append(self, item):
         schema = _get_schema_for_index(self._schema, len(self._instance))
         item = _cast(item, schema)
