@@ -1,9 +1,11 @@
+import contextlib
+import re
+
 import pytest
 from astropy.io import fits
 import numpy as np
 from numpy.testing import assert_array_almost_equal, assert_array_equal
 import asdf.schema
-from jsonschema import ValidationError
 
 from stdatamodels import DataModel
 from stdatamodels import fits_support
@@ -418,13 +420,19 @@ def test_skip_fits_update(tmp_path,
     with fits.open(file_path) as hduls:
         hduls[0].header['EXP_TYPE'] = 'FGS_DARK'
 
+        if skip_fits_update is not None:
+            ctx = pytest.warns(DeprecationWarning, match="skip_fits_update is deprecated")
+        else:
+            ctx = contextlib.nullcontext()
+
         if use_env:
             if skip_fits_update is not None:
                 monkeypatch.setenv("SKIP_FITS_UPDATE", str(skip_fits_update))
                 skip_fits_update = None
 
-        model = FitsModel(hduls, skip_fits_update=skip_fits_update)
-        assert model.meta.exposure.type == expected_exp_type
+        with ctx:
+            model = FitsModel(hduls, skip_fits_update=skip_fits_update)
+            assert model.meta.exposure.type == expected_exp_type
 
 
 def test_from_hdulist(tmp_path):
@@ -439,15 +447,6 @@ def test_from_hdulist(tmp_path):
         assert not hdulist.fileinfo(0)['file'].closed
 
 
-def test_open_fits_model_s3(s3_root_dir):
-    path = str(s3_root_dir.join("test.fits"))
-    with DataModel() as dm:
-        dm.save(path)
-
-    model = DataModel("s3://test-s3-data/test.fits")
-    assert isinstance(model, DataModel)
-
-
 def test_data_array(tmp_path):
     file_path = tmp_path/"test.fits"
     file_path2 = tmp_path/"test2.fits"
@@ -597,11 +596,6 @@ def test_ndarray_validation(tmp_path):
     with FitsModel(file_path, strict_validation=True, validate_arrays=True) as model:
         model.validate()
 
-    # But raise an error when casting is disabled
-    with pytest.raises(ValidationError, match="Array datatype 'float64' is not compatible with 'float32'"):
-        with FitsModel(file_path, strict_validation=True, cast_fits_arrays=False, validate_arrays=True) as model:
-            model.validate()
-
     # Wrong dimensions
     hdu = fits.ImageHDU(data=np.ones((4,), dtype=np.float64), name="SCI")
     hdul = fits.HDUList([fits.PrimaryHDU(), hdu])
@@ -612,11 +606,6 @@ def test_ndarray_validation(tmp_path):
         with FitsModel(file_path, strict_validation=True, validate_arrays=True) as model:
             model.validate()
 
-    # Should also be caught by validation
-    with pytest.raises(ValidationError, match="Wrong number of dimensions: Expected 2, got 1"):
-        with FitsModel(file_path, strict_validation=True, cast_fits_arrays=False, validate_arrays=True) as model:
-            model.validate()
-
 
 def test_resave_duplication_bug(tmp_path):
     """
@@ -636,3 +625,60 @@ def test_resave_duplication_bug(tmp_path):
 
     with fits.open(fn1) as ff1, fits.open(fn2) as ff2:
         assert ff1['ASDF'].size == ff2['ASDF'].size
+
+
+def test_table_linking(tmp_path):
+    file_path = tmp_path / "test.fits"
+
+    schema = {
+        'title': 'Test data model',
+        '$schema': 'http://stsci.edu/schemas/fits-schema/fits-schema',
+        'type': 'object',
+        'properties': {
+            'meta': {
+                'type': 'object',
+                'properties': {}
+            },
+            'test_table': {
+                'title': 'Test table',
+                'fits_hdu': 'TESTTABL',
+                'datatype': [
+                    {'name': 'A_COL', 'datatype': 'int8'},
+                    {'name': 'B_COL', 'datatype': 'int8'}
+                ]
+            }
+        }
+    }
+
+    with DataModel(schema=schema) as dm:
+        test_array = np.array([(1, 2), (3, 4)], dtype=[('A_COL', 'i1'), ('B_COL', 'i1')])
+
+        # assigning to the model will convert the array to a FITS_rec
+        dm.test_table = test_array
+        assert isinstance(dm.test_table, fits.FITS_rec)
+
+        # save the model (with the table)
+        dm.save(file_path)
+
+    # open the model and confirm that the table was linked to an hdu
+    with fits.open(file_path) as ff:
+        # read the bytes for the embedded ASDF content
+        asdf_bytes = ff['ASDF'].data.tobytes()
+
+        # get only the bytes for the tree (not blocks) by splitting
+        # on the yaml end document marker '...'
+        # on the first block magic sequence
+        tree_string = asdf_bytes.split(b'...')[0].decode('ascii')
+        unlinked_arrays = re.findall(r'source:\s+[^f]', tree_string)
+        assert not len(unlinked_arrays), unlinked_arrays
+
+
+def test_fitsrec_for_non_schema_data(tmp_path):
+    # make a file where some non-schema data is linked between
+    # the asdf extension and another hdu. This simulates an old
+    # file where this condition might occur (perhaps after a schema
+    # update).
+    m = DataModel()
+    m._instance['sneaky_table'] = fits.FITS_rec(np.array([('a', 1),], dtype=[('foo', 'S3'), ('bar', 'f4')]))
+    fn = tmp_path / "test.fits"
+    m.save(fn)
